from .base import Caption, CaptionConverter, CaptionList, CaptionNode, CaptionSet
from .dfxp import DFXPReader, DFXPWriter
from .exceptions import (
    CaptionLineLengthError,
    CaptionReadError,
    CaptionReadNoCaptions,
    CaptionReadSyntaxError,
)
from .microdvd import MicroDVDReader, MicroDVDWriter
from .sami import SAMIReader, SAMIWriter
from .scc import SCCReader, SCCWriter
from .scc.translator import translate_scc
from .srt import SRTReader, SRTWriter
from .transcript import TranscriptWriter
from .webvtt import WebVTTReader, WebVTTWriter

__all__ = [
    "CaptionConverter",
    "DFXPReader",
    "DFXPWriter",
    "MicroDVDReader",
    "MicroDVDWriter",
    "SAMIReader",
    "SAMIWriter",
    "SRTReader",
    "SRTWriter",
    "SCCReader",
    "SCCWriter",
    "translate_scc",
    "WebVTTReader",
    "WebVTTWriter",
    "CaptionReadError",
    "CaptionReadNoCaptions",
    "CaptionReadSyntaxError",
    "detect_format",
    "CaptionNode",
    "Caption",
    "CaptionList",
    "CaptionSet",
    "TranscriptWriter",
]

SUPPORTED_READERS = (
    DFXPReader,
    MicroDVDReader,
    WebVTTReader,
    SAMIReader,
    SRTReader,
    SCCReader,
)


def detect_format(caps):
    """
    Detect the format of the provided caption string.

    :returns: the reader class for the detected format.
    """
    if not len(caps):
        raise CaptionReadNoCaptions("Empty caption file")

    for reader in SUPPORTED_READERS:
        if reader().detect(caps):
            return reader

    return None



def is_leaf(element):
    """
    Return True if the element is a leaf, False otherwise. The element is
    considered a leaf if it is either NavigableString or the "br" tag
    :param element: A BeautifulSoup tag or NavigableString
    """
    name = getattr(element, "name", None)
    if not name or name == "br":
        return True
    return False



import re
from copy import deepcopy
from xml.sax.saxutils import escape

from bs4 import BeautifulSoup, NavigableString

from ..base import (
    DEFAULT_LANGUAGE_CODE,
    BaseReader,
    BaseWriter,
    Caption,
    CaptionList,
    CaptionNode,
    CaptionSet,
)
from ..exceptions import (
    CaptionReadNoCaptions,
    CaptionReadSyntaxError,
    CaptionReadTimingError,
    InvalidInputError,
)
from ..geometry import (
    Alignment,
    HorizontalAlignmentEnum,
    Layout,
    Padding,
    Point,
    Stretch,
    UnitEnum,
    VerticalAlignmentEnum,
)
from ..utils import is_leaf

__all__ = [
    "DFXP_BASE_MARKUP",
    "DFXP_DEFAULT_STYLE",
    "DFXP_DEFAULT_STYLE_ID",
    "DFXP_DEFAULT_REGION_ID",
    "DFXPReader",
    "DFXPWriter",
    "DFXP_DEFAULT_REGION",
]

DFXP_BASE_MARKUP = """
<tt xmlns="http://www.w3.org/ns/ttml"
    xmlns:tts="http://www.w3.org/ns/ttml#styling">
    <head>
        <styling/>
        <layout/>
    </head>
    <body/>
</tt>
"""

DFXP_DEFAULT_STYLE = {
    "color": "white",
    "font-family": "monospace",
    "font-size": "1c",
}

DFXP_DEFAULT_REGION = Layout(
    alignment=Alignment(HorizontalAlignmentEnum.START, VerticalAlignmentEnum.BOTTOM)
)

DFXP_DEFAULT_STYLE_ID = "default"
DFXP_DEFAULT_REGION_ID = "bottom"

CLOCK_TIME_PATTERN = (
    r"(?P<clock_time>(?P<hours>\d+):(?P<minutes>\d{2}):(?P<seconds>\d{2})"
    r"(:(?P<frames>\d{2})|\.(?P<sub_frames>\d+))?)"
)
OFFSET_TIME_PATTERN = (
    r"(?P<offset_time>(?P<time_count>\d+(\.\d+)?)" r"(?P<metric>h|m|s|ms|f|t))"
)
TIME_EXPRESSION_PATTERN = re.compile(rf"^({CLOCK_TIME_PATTERN}|{OFFSET_TIME_PATTERN})$")

MICROSECONDS_PER_UNIT = {
    "hours": 3600000000,
    "minutes": 60000000,
    "seconds": 1000000,
    "milliseconds": 1000,
}

DFXP_DEFAULT_LANGUAGE_CODE = "en"


class DFXPReader(BaseReader):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        """ This is to support positioning attributes directly on the paragraph
        elements. According to the documentation, these attributes shouldn't be
        taken into account:
        https://www.w3.org/TR/ttml1/#style-attribute-origin
        This attribute (origin, extent) may be specified by any element type 
        that permits use of attributes in the TT Style Namespace; however,
        this attribute applies as a style property only to those element types
        indicated in the following table."""
        self.read_invalid_positioning = kw.get("read_invalid_positioning", False)
        self.nodes = []

    def detect(self, content):
        if "</tt>" in content.lower():
            return True
        else:
            return False

    def read(self, content):
        if not isinstance(content, str):
            raise InvalidInputError("The content is not a unicode string.")

        dfxp_document = self._get_dfxp_parser_class()(
            content, read_invalid_positioning=self.read_invalid_positioning
        )

        caption_dict = {}
        style_dict = {}

        default_language = dfxp_document.tt.attrs.get("xml:lang", DEFAULT_LANGUAGE_CODE)

        # Each div represents all the captions for a single language.
        for div in dfxp_document.find_all("div"):
            lang = div.attrs.get("xml:lang", default_language)

            caption_dict[lang] = self._convert_div_to_caption_list(div)

        for style in dfxp_document.find_all("style"):
            id_ = style.attrs.get("xml:id") or style.attrs.get("id")
            if id_:
                # Don't create document styles for those styles that are
                # descendants of <region> tags. See link:
                # http://www.w3.org/TR/ttaf1-dfxp/#styling-vocabulary-style
                if "region" not in [parent_.name for parent_ in style.parents]:
                    style_dict[id_] = self._convert_style(style)

        caption_set = CaptionSet(caption_dict, styles=style_dict)

        if caption_set.is_empty():
            raise CaptionReadNoCaptions("empty caption file")

        return caption_set

    @staticmethod
    def _get_dfxp_parser_class():
        """Hook method for providing a custom DFXP parser"""
        return LayoutAwareDFXPParser

    def _convert_div_to_caption_list(self, div):
        return CaptionList(
            [
                caption
                for caption in (
                    self._convert_p_tag_to_caption(p_tag)
                    for p_tag in div.find_all("p")
                    if p_tag.get_text().strip()
                )
                if caption is not None
            ],
            div.layout_info,
        )

    def _convert_p_tag_to_caption(self, p_tag):
        start, end = self._find_and_convert_times(p_tag)
        self.nodes = []
        self._convert_tag_to_node(p_tag)
        styles = self._convert_style(p_tag)

        if len(self.nodes) > 0:
            return Caption(
                start, end, self.nodes, style=styles, layout_info=p_tag.layout_info
            )
        return None

    def _find_and_convert_times(self, p_tag):
        begin = p_tag.get("begin")
        if not begin:
            raise CaptionReadTimingError(f"Missing begin time on line {p_tag}.")

        end = p_tag.get("end")
        dur = p_tag.get("dur")
        if not end and not dur:
            raise CaptionReadTimingError(
                f"Missing end time or duration on line {p_tag}."
            )

        start = self._convert_timestamp_to_microseconds(begin)
        if end:
            end = self._convert_timestamp_to_microseconds(p_tag["end"])
        else:
            dur = self._convert_timestamp_to_microseconds(p_tag["dur"])
            end = start + dur

        return start, end

    def _convert_timestamp_to_microseconds(self, stamp):
        match = TIME_EXPRESSION_PATTERN.search(stamp)
        if not match:
            raise CaptionReadTimingError(
                f"Invalid timestamp: {stamp}. Accepted formats: hh:mm:ss / "
                "hh:mm:ss:ff / hh:mm:ss.sub-frames / time_count h|m|s|ms|f."
            )
        if match.group("clock_time"):
            return self._convert_clock_time_to_microseconds(match)
        else:
            return self._convert_time_count_to_microseconds(match)

    @staticmethod
    def _convert_clock_time_to_microseconds(clock_time_match):
        microseconds = (
            int(clock_time_match.group("hours")) * MICROSECONDS_PER_UNIT["hours"]
        )
        microseconds += (
            int(clock_time_match.group("minutes")) * MICROSECONDS_PER_UNIT["minutes"]
        )
        microseconds += (
            int(clock_time_match.group("seconds")) * MICROSECONDS_PER_UNIT["seconds"]
        )
        if clock_time_match.group("sub_frames"):
            microseconds += (
                int(clock_time_match.group("sub_frames").ljust(3, "0"))
                * MICROSECONDS_PER_UNIT["milliseconds"]
            )
        elif clock_time_match.group("frames"):
            microseconds += (
                int(clock_time_match.group("frames"))
                / 30
                * MICROSECONDS_PER_UNIT["seconds"]
            )
        return int(microseconds)

    @staticmethod
    def _convert_time_count_to_microseconds(time_count_match):
        value = float(time_count_match.group("time_count"))
        metric = time_count_match.group("metric")
        if metric == "h":
            microseconds = value * MICROSECONDS_PER_UNIT["hours"]
        elif metric == "m":
            microseconds = value * MICROSECONDS_PER_UNIT["minutes"]
        elif metric == "s":
            microseconds = value * MICROSECONDS_PER_UNIT["seconds"]
        elif metric == "ms":
            microseconds = value * MICROSECONDS_PER_UNIT["milliseconds"]
        elif metric == "f":
            microseconds = value / 30 * MICROSECONDS_PER_UNIT["seconds"]
        elif metric == "t":
            raise NotImplementedError(
                "The tick metric for time count is " "not currently implemented."
            )
        return int(microseconds)

    def _convert_tag_to_node(self, tag):
        # convert text
        if isinstance(tag, NavigableString):
            # strips indentation whitespace only
            pattern = re.compile("^(?:[\n\r]+\\s*)?(.+)")
            result = pattern.search(tag)
            if result:
                # Escaping/unescaping xml entities is the responsibility of the
                # xml parser used by BeautifulSoup in its initialization. The
                # content of the tag variable at this point should be a plain
                # unicode string with xml entities already converted to unicode
                # characters.
                tag_text = result.groups()[0]
                node = CaptionNode.create_text(tag_text, layout_info=tag.layout_info)
                self.nodes.append(node)
        # convert line breaks
        elif tag.name == "br":
            self.nodes.append(CaptionNode.create_break(layout_info=tag.layout_info))
        # convert italics
        elif tag.name == "span":
            # convert span
            self._convert_span_to_nodes(tag)
        else:
            # recursively call function for any children elements
            for a in tag.contents:
                self._convert_tag_to_node(a)

    def _convert_span_to_nodes(self, tag):
        # convert tag attributes
        args = self._convert_style(tag)
        # only include span tag if attributes returned
        # TODO - this is an obvious very old bug. args will be a dictionary.
        # but since nobody complained, I'll leave it like that.
        # Happy investigating!
        if args != "":
            node = CaptionNode.create_style(True, args, layout_info=tag.layout_info)
            node.start = True
            node.content = args
            self.nodes.append(node)

            # recursively call function for any children elements
            for a in tag.contents:
                self._convert_tag_to_node(a)
            node = CaptionNode.create_style(False, args, layout_info=tag.layout_info)
            node.start = False
            node.content = args
            self.nodes.append(node)
        else:
            for a in tag.contents:
                self._convert_tag_to_node(a)

    # convert style from DFXP
    def _convert_style(self, tag):
        """Converts the attributes of an XML node to a dictionary. This is a
         deprecated method of handling styling/ layout information, and
         overlaps (in partially known ways)with the newer way of doing stuff.

         For examples of how to refactor this, see the .layout_info attribute,
         and the geometry.Layout class.


        :param tag: BeautifulSoup Tag
        :rtype: dict
        """
        attrs = {}
        dfxp_attrs = tag.attrs
        for arg in dfxp_attrs:
            if arg.lower() == "style":
                # Support multiple classes per tag
                attrs["classes"] = dfxp_attrs[arg].strip().split(" ")
                # Save old class attribute for compatibility
                attrs["class"] = dfxp_attrs[arg]
            elif arg.lower() == "tts:fontstyle" and dfxp_attrs[arg] == "italic":
                attrs["italics"] = True
            elif arg.lower() == "tts:fontweight" and dfxp_attrs[arg] == "bold":
                attrs["bold"] = True
            elif (arg.lower() == "tts:textdecoration"
                  and "underline" in dfxp_attrs[arg].strip().split(" ")):
                attrs["underline"] = True
            elif arg.lower() == "tts:textalign":
                attrs["text-align"] = dfxp_attrs[arg]
            elif arg.lower() == "tts:fontfamily":
                attrs["font-family"] = dfxp_attrs[arg]
            elif arg.lower() == "tts:fontsize":
                attrs["font-size"] = dfxp_attrs[arg]
            elif arg.lower() == "tts:color":
                attrs["color"] = dfxp_attrs[arg]
        return attrs


class DFXPWriter(BaseWriter):
    def __init__(self, *args, **kwargs):
        self.write_inline_positioning = kwargs.pop("write_inline_positioning", False)
        self.p_style = False
        self.open_span = False
        self.region_creator = None
        super().__init__(*args, **kwargs)

    def write(self, caption_set, force=""):
        """Converts a CaptionSet into an equivalent corresponding DFXP file

        :type caption_set: pycaption.base.CaptionSet
        :param force: only use this language, if available in the
        caption_set

        :rtype: str
        """
        dfxp = BeautifulSoup(DFXP_BASE_MARKUP, "lxml-xml")

        langs = caption_set.get_languages()
        if force in langs:
            langs = [force]
            dfxp.find("tt")["xml:lang"] = force
        else:
            dfxp.find("tt")["xml:lang"] = DFXP_DEFAULT_LANGUAGE_CODE

        caption_set = deepcopy(caption_set)

        # Loop through all captions/nodes and apply transformations to layout
        # in function of the provided or default settings
        for lang in langs:
            for caption in caption_set.get_captions(lang):
                caption.layout_info = self._relativize_and_fit_to_screen(
                    caption.layout_info
                )
                for node in caption.nodes:
                    node.layout_info = self._relativize_and_fit_to_screen(
                        node.layout_info
                    )

        # Create the styles in the <styling> section, or a default style.
        for style_id, style in caption_set.get_styles():
            if style != {}:
                dfxp = self._recreate_styling_tag(style_id, style, dfxp)
        if not caption_set.get_styles():
            dfxp = self._recreate_styling_tag(
                DFXP_DEFAULT_STYLE_ID, DFXP_DEFAULT_STYLE, dfxp
            )

        self.region_creator = self._get_region_creator_class()(dfxp, caption_set)
        self.region_creator.create_document_regions()

        body = dfxp.find("body")

        for lang in langs:
            div = dfxp.new_tag("div")
            div["xml:lang"] = lang
            self._assign_positioning_data(div, lang, caption_set)

            for caption in caption_set.get_captions(lang):
                if caption.style:
                    caption_style = caption.style
                else:
                    caption_style = {"class": DFXP_DEFAULT_STYLE_ID}

                p = self._recreate_p_tag(
                    caption, caption_style, dfxp, caption_set, lang
                )
                self._assign_positioning_data(p, lang, caption_set, caption)
                div.append(p)

            body.append(div)
        self.region_creator.cleanup_regions()
        caption_content = dfxp.prettify(formatter=None)
        return caption_content

    @staticmethod
    def _get_region_creator_class():
        """Hook method for providing a custom RegionCreator"""
        return RegionCreator

    def _assign_positioning_data(
        self, tag, lang, caption_set=None, caption=None, caption_node=None
    ):
        """Modifies the current tag, assigning it the 'region' attribute.

        :param tag: the BeautifulSoup tag to be modified
        :type lang: str
        :param lang: the caption language
        :type caption_set: CaptionSet
        :param caption_set: The CaptionSet parent
        :type caption: Caption
        :type caption_node: CaptionNode
        """
        assigned_id, attribs = self.region_creator.get_positioning_info(
            lang, caption_set, caption, caption_node
        )

        if assigned_id:
            tag["region"] = assigned_id

            # Write non-standard positioning information
            if self.write_inline_positioning:
                tag.attrs.update(attribs)

    def _recreate_styling_tag(self, style, content, dfxp):
        # TODO - should be drastically simplified: if attributes : append
        dfxp_style = dfxp.new_tag("style")
        dfxp_style.attrs.update({"xml:id": style})

        attributes = _recreate_style(content, dfxp)
        dfxp_style.attrs.update(attributes)

        new_tag = dfxp.new_tag("style")
        new_tag.attrs.update({"xml:id": style})
        if dfxp_style != new_tag:
            dfxp.find("styling").append(dfxp_style)

        return dfxp

    def _recreate_p_tag(
        self, caption, caption_style, dfxp, caption_set=None, lang=None
    ):
        start = caption.format_start()
        end = caption.format_end()
        p = dfxp.new_tag("p", begin=start, end=end)
        p.string = self._recreate_text(caption, dfxp, caption_set, lang)

        if dfxp.find("style", {"xml:id": "p"}):
            p["style"] = "p"

        p.attrs.update(_recreate_style(caption_style, dfxp))

        return p

    def _recreate_text(self, caption, dfxp, caption_set=None, lang=None):
        line = ""

        for node in caption.nodes:
            if node.type_ == CaptionNode.TEXT:
                line += self._encode(node.content)

            elif node.type_ == CaptionNode.BREAK:
                line = line.rstrip() + "<br/>\n    "

            elif node.type_ == CaptionNode.STYLE:
                line = self._recreate_span(line, node, dfxp, caption_set, caption, lang)

        return line.rstrip()

    def _recreate_span(
        self, line, node, dfxp, caption_set=None, caption=None, lang=None
    ):
        # TODO - This method seriously has to go away!
        # Because of the CaptionNode.STYLE nodes, tree-like structures are
        # are really hard to build, and proper xml elements can't be added.
        # We are left with creating tags manually, which is hard to understand
        # and harder to maintain
        if node.start:
            styles = ""

            content_with_style = _recreate_style(node.content, dfxp)
            for style, value in list(content_with_style.items()):
                styles += f' {style}="{value}"'
            if node.layout_info:
                region_id, region_attribs = self.region_creator.get_positioning_info(
                    lang, caption_set, caption, node
                )
                styles += f' region="{region_id}"'
                if self.write_inline_positioning:
                    styles += " " + " ".join(
                        [f'{k_}="{v_}"' for k_, v_ in list(region_attribs.items())]
                    )

            if styles:
                if self.open_span:
                    line = line.rstrip() + "</span> "
                line += f"<span{styles}>"
                self.open_span = True

        elif self.open_span:
            line = line.rstrip() + "</span> "
            self.open_span = False

        return line

    def _encode(self, s):
        """
        Escapes XML 1.0 illegal or discouraged characters
        For details see:
            - http://www.w3.org/TR/2008/REC-xml-20081126/#dt-chardata
        :type s: str
        :param s: The content of a text node
        """
        return escape(s)


class LayoutAwareDFXPParser(BeautifulSoup):
    """This makes the xml instance capable of providing layout information
    for every one of its nodes (it adds a 'layout_info' attribute on each node)

    It parses the element tree in pre-order-like fashion as dictated by the
    dfxp specs here:
    http://www.w3.org/TR/ttaf1-dfxp/#semantics-style-resolution-process-overall

    TODO: Some sections require pre-order traversal, others post-order (e.g.
    http://www.w3.org/TR/ttaf1-dfxp/#semantics-region-layout-step-1). For the
    features we support, it was easier to use pre-order and it seems to have
    been enough. It should be clarified whether this is ok or not.
    """

    # A lot of elements will have no positioning info. Use this flyweight
    # to save memory
    NO_POSITIONING_INFO = None

    def __init__(
        self,
        markup="",
        features="html.parser",
        builder=None,
        parse_only=None,
        from_encoding=None,
        read_invalid_positioning=False,
        **kwargs,
    ):
        """The `features` param determines the parser to be used. The parsers
        are usually html parsers, some more forgiving than others, and as such
        they do stuff very differently especially for xml files. We chose this
        one because even though the docs say it's slower, it's very forgiving
        (it allows unescaped `<` characters, for example). It doesn't support
        the `&apos;` entity, however, since it respects the HTML4 and not HTML5
        syntax. Since this is valid XML 1.0, as a workaround we have to
        manually replace the every occurrence of this entity in the string
        before using the parser.

        The reason why we haven't used the 'xml' parser is that it destroys
        characters such as < or & (even the escaped ones).

        The 'lxml' parser seems to respect the html specification the best, but
        it's not as forgiving as 'html.parser' and fails when there are
        unescaped `<` characters in the input, for example.

        An alternative would be using html5lib, but that (1) is an external
        dependency and (2) BeautifulSoup says it's the slowest option.

        :type read_invalid_positioning: bool
        :param read_invalid_positioning: if True, will try to also look
            for layout info on every element itself (even if the docs explicitly
            call for ignoring attributes, when incorrectly placed)


        Check out the docs below for explanation.
        http://www.crummy.com/software/BeautifulSoup/bs4/doc/#installing-a-parser
        """

        # Work around for lack of '&apos;' support in html.parser
        markup = markup.replace("&apos;", "'")

        super().__init__(markup, features, builder, parse_only, from_encoding, **kwargs)

        self.read_invalid_positioning = read_invalid_positioning

        for div in self.find_all("div"):
            self._pre_order_visit(div)

    def _pre_order_visit(self, element, inherit_from=None):
        """Process the xml tree elements in pre order by adding a .layout_info
        attribute to each of them.

        The specs say this is how the attributes should be determined, but
        for the region attribute this might be irrelevant and any type of tree
        walk might do.
        :param element: a BeautifulSoup Tag or NavigableString.
        :param inherit_from: a Layout object with all the layout info
                inherited from the ancestors of the present node
        """
        if is_leaf(element):
            # The element is a leaf (e.g. NavigableString or <br>)
            element.layout_info = inherit_from
        else:
            region_id = self._determine_region_id(element)
            # TODO - this looks highly cachable. If it turns out too much
            # memory is being taken up by the caption set, cache this with a
            # WeakValueDict
            layout_info = self._extract_positioning_information(region_id, element)
            element.layout_info = layout_info
            for child in element.contents:
                self._pre_order_visit(child, inherit_from=layout_info)

    @staticmethod
    def _get_region_from_ancestors(element):
        """Try to get the region ID from the nearest ancestor that has it"""
        region_id = None
        parent = element.parent
        while parent:
            region_id = parent.get("region")
            if region_id:
                break
            parent = parent.parent

        return region_id

    @staticmethod
    def _get_region_from_descendants(element):
        """Try to get the region_id from the closest descendant (that has it)
        This is trickier, because at different times, the determined region
        could be different. If this happens, discard region data
        """
        # element might be a NavigableString, not a Tag.
        # if is_leaf(element):
        if isinstance(element, NavigableString):
            return None

        region_id = None
        child_region_ids = {child.get("region") for child in element.findChildren()}
        if len(child_region_ids) > 1:
            raise LookupError
        if len(child_region_ids) == 1:
            region_id = child_region_ids.pop()

        return region_id

    @classmethod
    def _determine_region_id(cls, element):
        """Determines the TT region of an element.

        For determining the region of an element, check out the url, look for
        section "[associate region]". One difference, is that we leave the
        default region id empty. The writer will know what to do:
        http://www.w3.org/TR/ttaf1-dfxp/#semantics-region-layout-step-1

        :param element: the xml element for which we're trying to get region
            info
        """
        # element could be a NavigableString. Those are dumb.
        region_id = None

        if hasattr(element, "get"):
            region_id = element.get("region")

        if not region_id:
            region_id = cls._get_region_from_ancestors(element)

        if not region_id:
            try:
                region_id = cls._get_region_from_descendants(element)
            except LookupError:
                return

        return region_id

    def _extract_positioning_information(self, region_id, element):
        """Returns a Layout object that describes the element's positioning
        information

        :param region_id: the id of the region to which the element is
            associated
        :type region_id: str
        :param element: BeautifulSoup Tag or NavigableString; this only comes
            into action (at the moment) if the
        :rtype: Layout
        """
        region_tag = None

        if region_id is not None:
            region_tag = self.find("region", {"xml:id": region_id})

        region_scraper = self._get_layout_info_scraper_class()(self, region_tag)

        layout_info = region_scraper.scrape_positioning_info(
            element, self.read_invalid_positioning
        )

        if layout_info and any(layout_info):
            # layout_info contains information?
            return self._get_layout_class()(*layout_info)
        else:
            # layout_info doesn't contain any information
            return self.NO_POSITIONING_INFO

    @staticmethod
    def _get_layout_info_scraper_class():
        """Hook method for getting an implementation of a LayoutInfoScraper."""
        return LayoutInfoScraper

    @staticmethod
    def _get_layout_class():
        """Hook method for providing the Layout class to use"""
        return Layout


class LayoutInfoScraper:
    """Encapsulates the methods for determining the layout information about
    an element (with the element's region playing an important role).
    """

    def __init__(self, document, region=None):
        """
        :param document: the BeautifulSoup document instance, of which `region`
            is a descendant
        :param region: the region tag
        """
        self.region = region
        self._styling_section = document.findChild("styling")
        if region:
            self.region_styles = self._get_style_sources(self._styling_section, region)
        else:
            self.region_styles = []
        self.root_element = document.find("tt")

    @classmethod
    def _get_style_sources(cls, styling_section, element):
        """Returns a list, containing  tags, in the order they should be
        evaluated, for determining layout information.

        This method should be extended if the styles provided by it are not
        enough (like for the captions created with CaptionMaker 6, which are
        not compatible with the specs)

        Check the URL for detailed description of how styles should be resolved
        http://www.w3.org/TR/ttaf1-dfxp/#semantics-style-association

        Returns:
          1. All child styles of the element, each with its reference chain
          2. The style referenced by the element, via the attrib. style="asdf"
            together with its reference chain
        Note: the specs are unclear about the priority of styles that are
        referenced by nested styles. I've assumed it's higher than referential
        styling
        """
        # If we're analyzing a NavigableString, just quit
        if not hasattr(element, "findAll"):
            return ()

        nested_styles = []

        # <div> tags have a huge number of children, with highly unlikely
        # <style> tags among them. Looping through all of them as is, would
        # make the DFXP parser freeze for a long time, so we skip this step
        # if the parent is a <div> tag. Technically, this step shouldn't be
        # skipped, but it would make the reader read in O(n^2) (half an hour
        # for 1500 timed captions)
        if element.name not in ("div", "body", "tt"):
            for style in element.contents:
                if getattr(style, "name", None) == "style":
                    nested_styles.extend(
                        cls._get_style_reference_chain(style, styling_section)
                    )

        referenced_style_id = element.get("style")

        referenced_styles = []
        if referenced_style_id and styling_section:
            referenced_style = styling_section.findChild(
                "style", {"xml:id": referenced_style_id}
            )

            referenced_styles = cls._get_style_reference_chain(
                referenced_style, styling_section
            )
        return nested_styles + referenced_styles

    @classmethod
    def _get_style_reference_chain(cls, style, styling_tag):
        """If style s1 references s2, and s3 -> s4 -> s5 -> ... -> sn,
        if called with s1, this returns [s1, s2, ... sn] (supposing all the
        styles are defined in the styling section, or stops at the last found
        style)

        :param style: a style tag, that might refer another style
        :param styling_tag: The tag representing the '<styling>' section of the
            dfxp document
        """
        if not style:
            return []

        result = [style]

        if not styling_tag:
            return result

        reference = style.get("style")

        if reference:
            referenced_styles = styling_tag.findChildren("style", {"xml:id": reference})

            if len(referenced_styles) == 1:
                return result + cls._get_style_reference_chain(
                    referenced_styles[0], styling_tag
                )
            elif len(referenced_styles) > 1:
                raise CaptionReadSyntaxError(
                    "Invalid caption file. "
                    f"More than 1 style with 'xml:id': {reference}"
                )

        return result

    def scrape_positioning_info(self, element=None, even_invalid=False):
        """Determines the positioning information tuple
        (origin, extent, padding, alignment) from the region element.

        The first 3 attributes can be specified inline, on the region node,
        on child tags of type <style> or on referenced <style> tags.

        The fourth attribute can be specified like the first 3, or on the xml
        element itself.

        If the attributes can't be determined, default values are returned
        (where such values exist) - XXX this is incorrect. No default values
        should be provided.

        :param element: BeautifulSoup Tag or NavigableString
        :type even_invalid: bool
        :param even_invalid: if True, will search attributes on the element
            (when they really should be checked only on the region)
        :rtype: tuple
        """
        usable_elem = element if even_invalid else None

        origin = (
            self._find_attribute(
                usable_elem, "tts:origin", Point.from_xml_attribute, ["auto"]
            )
            or DFXP_DEFAULT_REGION.origin
        )

        extent = self._find_attribute(
            usable_elem, "tts:extent", Stretch.from_xml_attribute, ["auto"]
        )

        if not extent:
            extent = self._find_root_extent() or DFXP_DEFAULT_REGION.extent

        padding = (
            self._find_attribute(usable_elem, "tts:padding", Padding.from_xml_attribute)
            or DFXP_DEFAULT_REGION.padding
        )

        # tts:textAlign is a special attribute, which can not be ignored when
        # specified on the element itself (only <p> nodes matter)
        # On elements like <span> it is also read, because this was legacy
        # behavior.
        if getattr(element, "name", None) in ("span", "p"):
            text_align_source = element
        else:
            text_align_source = None

        text_align = self._find_attribute(
            text_align_source, "tts:textAlign"
        ) or _create_external_horizontal_alignment(
            DFXP_DEFAULT_REGION.alignment.horizontal
        )
        display_align = self._find_attribute(
            usable_elem, "tts:displayAlign"
        ) or _create_external_vertical_alignment(DFXP_DEFAULT_REGION.alignment.vertical)
        alignment = _create_internal_alignment(text_align, display_align)

        return origin, extent, padding, alignment

    def _find_attribute_on_element_or_styles(
        self, attribute_name, element, factory, ignore, ignorecase
    ):
        """Look up the given attribute on the element, and all the styles
        referenced by it.

        :type attribute_name: str
        :param element: BeautifulSoup Tag or NavigableString
        :param factory: a function, to apply to the xml attribute
        :param ignore: a list of values to ignore
        :type ignore: list
        :param ignorecase: Whether to ignore the casing
        :type ignorecase: bool
        :return: The result of applying the `factory` to the found attribute
            value, or None
        """
        value = _get_object_from_attribute(
            element, attribute_name, factory, ignore, ignorecase
        )
        if value is None:
            # Does a referenced style of the element have it?
            for style in self._get_style_sources(self._styling_section, element):
                value = _get_object_from_attribute(
                    style, attribute_name, factory, ignore, ignorecase
                )
                if value:
                    break
        return value

    def _find_attribute(
        self, element, attribute_name, factory=lambda x: x, ignore=(), ignorecase=True
    ):
        """Try to find the `attribute_name` specified on the element, all its
        parents and all their styles (and referenced styles).

        :param element: BeautifulSoup Tag or NavigableString
        :type attribute_name: str
        :param attribute_name: the name of the attribute to resolve
        :type attribute_name: str
        :param factory: callable to transform the xml attribute into something
        :param ignore: iterable of values to ignore (will return None if the
            xml attribute is in that list)
        :param ignorecase: if True, the attribute will be searched in lowercase
            too
        :type ignorecase: bool
        :rtype: str
        :raises CaptionSyntaxError:
        """
        value = None

        # Does the element itself have it inline, or any of its styles?
        if element:
            value = self._find_attribute_on_element_or_styles(
                attribute_name, element, factory, ignore, ignorecase
            )

            if value is None:
                # Do any of the element's parents have the attribute?
                for parent in element.parents:
                    value = self._find_attribute_on_element_or_styles(
                        attribute_name, parent, factory, ignore, ignorecase
                    )
                    if value:
                        break

        # Does self.region or any of its styles have it?
        if value is None:
            value = self._find_attribute_on_element_or_styles(
                attribute_name, self.region, factory, ignore, ignorecase
            )

        return value

    def _find_root_extent(self):
        """Finds the "tts:extent" for the root <tt> element

        The tts:extent attribute, like the "tts:origin", can be specified on
        the region, its styles, or can be inherited from the root <tt> element.
        For the latter case, it must be specified in the unit 'pixel'.

        :rtype: Stretch
        """
        extent = None

        # Does the root 'tt' element have it?
        if extent is None:
            root = self.root_element
            extent = _get_object_from_attribute(
                root, "tts:extent", Stretch.from_xml_attribute
            )

            if extent is not None:
                if not extent.is_measured_in(UnitEnum.PIXEL):
                    raise CaptionReadSyntaxError(
                        "The base <tt> element attribute 'tts:extent' should "
                        "only be specified in pixels. Check the docs: "
                        "http://www.w3.org/TR/ttaf1-dfxp/"
                        "#style-attribute-extent"
                    )
        return extent


class RegionCreator:
    """Creates the DFXP regions, and knows how retrieve them, for assigning
    region IDs to every element

    # todo - needs to remember the IDs created, and later, when assigning a
    region to every dfxp element, needs to know what region to assign to that
    element, based on the CaptionNode, its Caption and its CaptionSet.

    The layout information for a node is determined like this:
        - If a node has a (NON-NULL*).layout_info attribute, return the region
            created for that exact specification
        - If a node has .layout_info = NULL*, retrieve the .layout_info from
            its Caption parent... if still NULL*, retrieve it from its
            CaptionSet
        - If the retrieval still resulted in None, assign to it the Default
            region

        *: NULL means LayoutAwareBeautifulParser.NO_POSITIONING_INFO
    """

    def __init__(self, dfxp, caption_set):
        """
        :type dfxp: BeautifulSoup
        :type caption_set: CaptionSet
        """
        self._dfxp = dfxp
        self._caption_set = caption_set
        self._region_map = {}
        self._id_seed = 0
        self._assigned_region_ids = set()

    @staticmethod
    def _collect_unique_regions(caption_set, ignore_region):
        """Iterate through all the nodes in the caption set, and return a list
        of all unique region specs (Layout objects)

        If a default region was created, and any scraped region matches its
        attributes, don't duplicate the region (eliminate that region from the
        result set)

        :type caption_set: CaptionSet
        :return: iterable containing the unique regions that will have to
            appear in the document
        """
        # This used to be a set, however since set order depends on the hash,
        # this messed up the tests every time some little detail was added to
        # the Layout class, or its references (which is highly fragile)
        unique_regions = _OrderedSet()
        # Get all the regions for all the <div>'s..corresponding to all the
        # languages
        languages = caption_set.get_languages()
        for lang in languages:
            layout_info = caption_set.get_layout_info(lang)
            unique_regions.add(layout_info)

            # Get the regions of all the captions.. (the <p> tags)
            for caption in caption_set.get_captions(lang):
                unique_regions.add(caption.layout_info)

                # The regions of all the text/br/style nodes
                for node in caption.nodes:
                    unique_regions.add(node.layout_info)

        unique_regions.discard(None)
        unique_regions.discard(ignore_region)
        return unique_regions

    @staticmethod
    def _create_unique_regions(unique_layouts, dfxp, id_factory):
        """Create each one of the regions in the list, inside the dfxp
        document, under the 'layout' section.

        :param unique_layouts: an iterable (unique!) geometry.Layout instances,
            describing the properties to be added to the dfxp regions
        :type dfxp: BeautifulSoup
        :param id_factory: A callable which generates unique IDs
        :return: a dict, mapping each unique layout to the ID of the region
            created for it
        :rtype: dict
        """
        region_map = {}
        layout_section = dfxp.find("layout")

        for region_spec in unique_layouts:
            if (
                region_spec.origin
                or region_spec.extent
                or region_spec.padding
                or region_spec.alignment
            ):
                new_region = dfxp.new_tag("region")
                new_id = id_factory()
                new_region["xml:id"] = new_id

                region_map[region_spec] = new_id
                region_attribs = _convert_layout_to_attributes(region_spec)
                new_region.attrs.update(region_attribs)

                layout_section.append(new_region)
        return region_map

    def create_document_regions(self):
        """Create the <region> tags required to position all the captions.

        Makes sure we have a default region
        """
        # Creates the default region
        default_region_map = self._create_unique_regions(
            [DFXP_DEFAULT_REGION], self._dfxp, lambda: DFXP_DEFAULT_REGION_ID
        )
        unique_regions = self._collect_unique_regions(
            self._caption_set, DFXP_DEFAULT_REGION
        )

        # Create the document specified regions
        self._region_map = self._create_unique_regions(
            unique_regions, self._dfxp, self._get_new_id
        )

        self._region_map.update(default_region_map)

    def _get_new_id(self, prefix="r"):
        """Return new, unique ids (use an internal counter).

        :type prefix: str
        """
        new_id = f"{prefix}{self._id_seed}"
        self._id_seed += 1
        return new_id

    def get_positioning_info(
        self, lang, caption_set=None, caption=None, caption_node=None
    ):
        """For the given element will return a valid region ID, used for
        assigning to the element, and a dict containing the positioning
        attributes of that region (useful for inline non-standard positioning)

        For the region_id to be returned for the entire CaptionSet, don't
        supply the `caption` or `caption_node` params.

        For the region_id to be returned for the Caption, don't supply the
        `caption_node` param

        <div> tags mean the caption is None and caption_node is None.
        <p> tags mean the caption_node is None

        :type lang: str
        :param lang: the language of the current caption element
        :type caption_set: CaptionSet
        :type caption: Caption
        :type caption_node: CaptionNode
        :rtype: tuple
        :return: (str, dict)
        """
        # More intelligent people would have used an elem.parent.parent..parent
        # pattern, but pycaption is not yet structured for this. 3 params
        # is not too much of a bother. If someone wants to make the structure
        # tree-like, they can easily change this.
        layout_info = None
        if caption_node:
            layout_info = caption_node.layout_info

        if not layout_info and caption:
            layout_info = caption.layout_info

        if not layout_info and caption_set:
            layout_info = caption_set.get_layout_info(lang)
            if not layout_info:
                layout_info = caption_set.layout_info

        region_id = self._region_map.get(layout_info)

        # Make sure the default region ID/ attributes are always returned
        # as fallback
        if not region_id:
            region_id = DFXP_DEFAULT_REGION_ID

        positioning_attributes = _convert_layout_to_attributes(layout_info)

        # Mark the region as having been assigned, so we can perform cleanup
        self._assigned_region_ids.add(region_id)

        return region_id, positioning_attributes

    def cleanup_regions(self):
        """Remove the unused regions from the output file"""
        layout_tag = self._dfxp.find("layout")
        if not layout_tag:
            return

        regions = layout_tag.findChildren("region")
        if not regions:
            return

        for region in regions:
            if region.attrs.get("xml:id") not in self._assigned_region_ids:
                region.extract()


def _recreate_style(content, dfxp):
    dfxp_style = {}

    if "class" in content:
        if dfxp.find("style", {"xml:id": content["class"]}):
            dfxp_style["style"] = content["class"]
    if "text-align" in content:
        dfxp_style["tts:textAlign"] = content["text-align"]
    if "italics" in content:
        dfxp_style["tts:fontStyle"] = "italic"
    if "font-family" in content:
        dfxp_style["tts:fontFamily"] = content["font-family"]
    if "font-size" in content:
        dfxp_style["tts:fontSize"] = content["font-size"]
    if "color" in content:
        dfxp_style["tts:color"] = content["color"]
    if "display-align" in content:
        dfxp_style["tts:displayAlign"] = content["display-align"]

    return dfxp_style


# TODO - highly cacheable. use WeakValueDict to improve performance
def _create_internal_alignment(text_align, display_align):
    """Given the 2 DFXP specific attributes, return the internal representation
    of an alignment

    In DFXP, the tts:textAlign can have the values
        "left", "center", "right", "start" and "end"
        with the default being "start".
    We interpret "start" as "left"... we don't yet support languages
    with right-to-left writing

    The "tts:displayAlign" can have the values
        "before", "center" and "after",
    with the default of "before". These refer to top/bottom positioning.

    :type text_align: str
    :type display_align: str
    :rtype: Alignment
    """
    if not (text_align or display_align):
        return None

    return Alignment.from_horizontal_and_vertical_align(text_align, display_align)


def _create_external_horizontal_alignment(horizontal_component):
    """From an internal horizontal alignment value, create a value to be used
    in the dfxp output file.

    :type horizontal_component: str
    :rtype: str
    """
    result = None

    if horizontal_component == HorizontalAlignmentEnum.LEFT:
        result = "left"
    if horizontal_component == HorizontalAlignmentEnum.CENTER:
        result = "center"
    if horizontal_component == HorizontalAlignmentEnum.RIGHT:
        result = "right"
    if horizontal_component == HorizontalAlignmentEnum.START:
        result = "start"
    if horizontal_component == HorizontalAlignmentEnum.END:
        result = "end"

    return result


def _create_external_vertical_alignment(vertical_component):
    """Given an alignment value used in the internal representation of the
    caption, return a value usable in the actual dfxp output file.

    :type vertical_component: str
    :rtype: str
    """
    result = None

    if vertical_component == VerticalAlignmentEnum.TOP:
        result = "before"
    if vertical_component == VerticalAlignmentEnum.CENTER:
        result = "center"
    if vertical_component == VerticalAlignmentEnum.BOTTOM:
        result = "after"

    return result


def _create_external_alignment(alignment):
    """Given an alignment object, return a dictionary. The keys are the dfxp
    attributes, and the value the dfxp values for the 'tts:textAlign' and
    'tts:displayAlign' attributes

    :type alignment: Alignment
    :rtype: dict
    """
    result = {}
    if not alignment:
        return result

    if not (alignment.horizontal or alignment.vertical):
        return result

    horizontal_alignment = _create_external_horizontal_alignment(alignment.horizontal)
    if horizontal_alignment:
        result["tts:textAlign"] = horizontal_alignment

    vertical_alignment = _create_external_vertical_alignment(alignment.vertical)
    if vertical_alignment:
        result["tts:displayAlign"] = vertical_alignment

    return result


def _get_object_from_attribute(
    tag, attr_name, factory, ignore_vals=(), ignorecase=True
):
    """For the xml `tag`, tries to retrieve the attribute `attr_name` and
    pass that to the factory in order to get a result. If the value of the
    attribute is in the `ignore_vals` iterable, returns None.

    :param tag: a BeautifulSoup tag
    :param attr_name: a string; represents an xml attribute name
    :param factory: a callable to transform the attribute into something
        usable (such as the classes from .geometry)
    :param ignore_vals: iterable of attribute values to ignore
    :raise CaptionReadSyntaxError: if the attribute has some crazy value
    """
    if not hasattr(tag, "has_attr"):
        return

    attr_value = None
    if tag.has_attr(attr_name):
        attr_value = tag.get(attr_name)

    if ignorecase and attr_name is not None:
        attr_value = tag.get(attr_name.lower())

    if attr_value is None:
        return

    usable_value = None

    if attr_value not in ignore_vals:
        try:
            usable_value = factory(attr_value)
        except ValueError as err:
            raise CaptionReadSyntaxError(err)

    return usable_value


def _convert_layout_to_attributes(layout):
    """Takes a layout object, and returns a dict whose keys are the dfxp
    attribute names, and the values are the dfxp attr. values.

    If the layout is None, return region default attributes

    :type layout: Layout
    :rtype: dict
    """
    result = {}
    if not layout:
        return _create_external_alignment(DFXP_DEFAULT_REGION.alignment)

    if layout.origin:
        result["tts:origin"] = layout.origin.to_xml_attribute()

    if layout.extent:
        result["tts:extent"] = layout.extent.to_xml_attribute()

    if layout.padding:
        result["tts:padding"] = layout.padding.to_xml_attribute()

    if layout.alignment:
        result.update(_create_external_alignment(layout.alignment))
    else:
        result.update(_create_external_alignment(DFXP_DEFAULT_REGION.alignment))

    return result


class _OrderedSet(list):
    """Quick implementation of a set that tracks the order. If this is a
    performance bottleneck, replace it with some other implementation.
    """

    def add(self, p_object):
        if p_object not in self:
            super().append(p_object)

    def discard(self, value):
        if value in self:
            super().remove(value)



import datetime
import re
import sys
from copy import deepcopy

from .base import BaseReader, BaseWriter, Caption, CaptionList, CaptionNode, CaptionSet
from .exceptions import (
    CaptionReadError,
    CaptionReadNoCaptions,
    CaptionReadSyntaxError,
    InvalidInputError,
)
from .geometry import HorizontalAlignmentEnum, Layout, Point, Size, Stretch, UnitEnum

# A WebVTT timing line has both start/end times and layout related settings
# (referred to as 'cue settings' in the documentation)
# The following pattern captures [start], [end] and [cue settings] if existent
TIMING_LINE_PATTERN = re.compile(r"^(\S+)\s+-->\s+(\S+)(?:\s+(.*?))?\s*$")
TIMESTAMP_PATTERN = re.compile(r"^(\d+):(\d{2})(:\d{2})?\.(\d{3})")
VOICE_SPAN_PATTERN = re.compile("<v(\\.\\w+)* ([^>]*)>")
OTHER_SPAN_PATTERN = re.compile(
    r"</?([cibuv]|ruby|rt|lang|(\d+):(\d{2})(:\d{2})?\.(\d{3})).*?>"
)  # These WebVTT tags are stripped off the cues on conversion
REGION_SETTING_PATTERN = re.compile(r"^([\w]+):(.+)$")
"""  
Matches a setting name (word chars) followed by colon and a value:
id:region1
width:50%
"""
REGION_ANCHOR_PATTERN = re.compile(r"^(\d+(?:\.\d+)?)%,(\d+(?:\.\d+)?)%$")
"""
Matches two percentage values (integer or decimal) separated by a comma:
0%,0%
100%,100%
"""
LINE_HEIGHT_VH = 5.33

WEBVTT_VERSION_OF = {
    HorizontalAlignmentEnum.LEFT: "left",
    HorizontalAlignmentEnum.CENTER: "center",
    HorizontalAlignmentEnum.RIGHT: "right",
    HorizontalAlignmentEnum.START: "start",
    HorizontalAlignmentEnum.END: "end",
}

DEFAULT_ALIGN = "start"


def microseconds(h, m, s, f):
    """
    Returns an integer representing a number of microseconds
    :rtype: int
    """
    return (int(h) * 3600 + int(m) * 60 + int(s)) * 1000000 + int(f) * 1000


class WebVTTReader(BaseReader):
    def __init__(
        self, ignore_timing_errors=True, time_shift_milliseconds=0, *args, **kwargs
    ):
        """
        :param ignore_timing_errors: Whether to ignore timing checks
        :type ignore_timing_errors: bool
        :param time_shift_milliseconds: Move all the timestamps forward/backward with this number of milliseconds
        :type time_shift_milliseconds: int
        """
        self.ignore_timing_errors = ignore_timing_errors
        self.time_shift_microseconds = time_shift_milliseconds * 1000

    def detect(self, content):
        return "WEBVTT" in content

    def read(self, content, lang="en-US"):
        if not isinstance(content, str):
            raise InvalidInputError("The content is not a unicode string.")

        caption_set = CaptionSet({lang: self._parse(content.splitlines())})

        if caption_set.is_empty():
            raise CaptionReadNoCaptions("empty caption file")

        return caption_set

    def _parse(self, lines):
        # State machine: cycles through waiting-for-timing → collecting-text
        # → emit-caption-on-blank-line, repeat.
        captions = CaptionList()
        start = None
        end = None
        nodes = []
        layout_info = None
        found_timing = False

        # Parse REGION blocks from the header area before processing cues
        self._regions = self._parse_regions(lines)

        for i, line in enumerate(lines):
            if "-->" in line:
                # Timing line found (e.g. "00:00:01.000 --> 00:00:03.000")
                # marks the start of a new cue
                found_timing = True
                timing_line = i
                last_start_time = captions[-1].start if captions else 0
                try:
                    start, end, layout_info = self._parse_timing_line(
                        line, last_start_time
                    )
                except CaptionReadError as e:
                    new_msg = f"{e.args[0]} (line {timing_line})"
                    tb = sys.exc_info()[2]
                    raise type(e)(new_msg).with_traceback(tb) from None

            elif "" == line:
                # Blank line = block separator in WebVTT.
                # If we were collecting a cue, finalize and store it.
                if found_timing and nodes:
                    found_timing = False
                    caption = Caption(start, end, nodes, layout_info=layout_info)
                    captions.append(caption)
                    nodes = []
            else:
                if found_timing:
                    # We're inside a cue — this line is cue text.
                    # Add a line break between multi-line cue text.
                    if nodes:
                        nodes.append(CaptionNode.create_break())
                    nodes.append(CaptionNode.create_text(self._decode(line)))
                else:
                    # Outside a cue: cue identifiers, NOTE blocks,
                    # or other metadata — skip silently.
                    pass

        # File may not end with a blank line — emit any remaining cue
        if nodes:
            caption = Caption(start, end, nodes, layout_info=layout_info)
            captions.append(caption)

        return captions

    def _validate_timings(self, start, end, last_start_time):
        if start is None:
            raise CaptionReadSyntaxError("Invalid cue start timestamp.")
        if end is None:
            raise CaptionReadSyntaxError("Invalid cue end timestamp.")
        if start > end:
            raise CaptionReadError("End timestamp is not greater than start timestamp.")
        if start < last_start_time:
            raise CaptionReadError(
                "Start timestamp is not greater than or equal"
                "to start timestamp of previous cue."
            )

    def _parse_timing_line(self, line, last_start_time):
        """
        :returns: Tuple (int, int, Layout)
        """
        m = TIMING_LINE_PATTERN.search(line)
        if not m:
            raise CaptionReadSyntaxError("Invalid timing format.")

        start = self._parse_timestamp(m.group(1)) + self.time_shift_microseconds
        end = self._parse_timestamp(m.group(2)) + self.time_shift_microseconds

        cue_settings = m.group(3)

        if not self.ignore_timing_errors:
            self._validate_timings(start, end, last_start_time)

        layout_info = None
        if cue_settings:
            region_id = self._extract_region_id(cue_settings)
            if region_id and region_id in self._regions:
                layout_info = self._regions[region_id]
                layout_info = Layout(
                    origin=layout_info.origin,
                    extent=layout_info.extent,
                    padding=layout_info.padding,
                    alignment=layout_info.alignment,
                    webvtt_positioning=cue_settings,
                )
            else:
                layout_info = Layout(webvtt_positioning=cue_settings)

        return start, end, layout_info

    def _parse_timestamp(self, timestamp):
        """Returns an integer representing a number of microseconds
        :rtype: int
        """
        m = TIMESTAMP_PATTERN.search(timestamp)
        if not m:
            raise CaptionReadSyntaxError("Invalid timing format.")

        m = m.groups()

        if m[2]:
            # Timestamp of the form [hours]:[minutes]:[seconds].[milliseconds]
            return microseconds(m[0], m[1], m[2].replace(":", ""), m[3])
        else:
            # Timestamp of the form [minutes]:[seconds].[milliseconds]
            return microseconds(0, m[0], m[1], m[3])

    def _decode(self, s):
        """
        Convert cue text from WebVTT XML-like format to plain unicode.
        :type s: str
        """
        s = s.strip()
        # Covert voice span
        s = VOICE_SPAN_PATTERN.sub("\\2: ", s)
        # TODO: Add support for other WebVTT tags. For now just strip them
        # off the text.
        s = OTHER_SPAN_PATTERN.sub("", s)
        # Replace WebVTT special XML codes with plain unicode values
        s = s.replace("&lt;", "<")
        s = s.replace("&gt;", ">")
        s = s.replace("&lrm;", "\u200e")
        s = s.replace("&rlm;", "\u200f")
        s = s.replace("&nbsp;", "\u00a0")
        # Must do ampersand last
        s = s.replace("&amp;", "&")
        return s

    def _parse_regions(self, lines):
        """Parse REGION blocks from the file header area.

        A WebVTT region defines a named rectangular area on screen where cues
        can be rendered. Regions appear before any cues with the syntax:

            REGION
            id:subtitle_area
            width:50%
            lines:3
            regionanchor:0%,100%
            viewportanchor:10%,90%
            scroll:up

        Supported settings:
            id             - unique identifier (required)
            width          - region width as percentage (default: 100%)
            lines          - visible line count (default: 3)
            regionanchor   - anchor point within region as x%,y% (default: 0%,100%)
            viewportanchor - anchor point on viewport as x%,y% (default: 0%,100%)
            scroll         - scroll behavior, only "up" is valid (default: none)

        :returns: dict mapping region id -> Layout
        """
        regions = {}
        i = 0
        while i < len(lines):
            line = lines[i].strip()
            if line == "REGION" or line.startswith(("REGION\t", "REGION ")):
                i += 1
                settings = {}
                seen_keys = set()
                # Read settings until a blank line (block separator in WebVTT)
                while i < len(lines) and lines[i].strip() != "":
                    # Match key:value pair (e.g. "width:50%")
                    m = REGION_SETTING_PATTERN.match(lines[i].strip())
                    if m:
                        key, value = m.group(1), m.group(2)
                        if key in seen_keys:
                            i += 1
                            continue
                        seen_keys.add(key)
                        settings[key] = value
                    i += 1
                # Skip regions without id (spec requires it; cues can't reference them)
                if "id" in settings:
                    region_id = settings["id"]
                    # First definition wins; duplicates are ignored (RULE-REG-009)
                    if region_id not in regions:
                        regions[region_id] = self._region_to_layout(settings)
            elif "-->" in line:
                # REGIONs only appear before cues; stop scanning once cues begin
                break
            else:
                i += 1
        return regions

    def _region_to_layout(self, settings):
        """Convert parsed region settings dict into a Layout with origin/extent.

        Uses W3C TTML-WebVTT mapping formulas:
            origin_x = viewportanchor_x - (regionanchor_x / 100 * width)
            origin_y = viewportanchor_y - (regionanchor_y / 100 * height)
            height = lines * 5.33
        """
        # Spec defaults per W3C WebVTT §6
        width = 100.0
        lines = 3
        regionanchor_x, regionanchor_y = 0.0, 100.0
        viewportanchor_x, viewportanchor_y = 0.0, 100.0
        # Parse each setting, falling back to defaults on invalid values
        if "width" in settings:
            try:
                width = float(settings["width"].rstrip("%"))
            except ValueError:
                pass

        if "lines" in settings:
            try:
                lines = int(settings["lines"])
            except ValueError:
                pass

        # regionanchor: which point inside the region is "pinned"
        # e.g. 0%,100% means the bottom-left corner of the region
        if "regionanchor" in settings:
            m = REGION_ANCHOR_PATTERN.match(settings["regionanchor"])
            if m:
                regionanchor_x = float(m.group(1))
                regionanchor_y = float(m.group(2))

        # viewportanchor: where on the screen that pin is placed
        # e.g. 10%,90% means 10% from left, 90% from top
        if "viewportanchor" in settings:
            m = REGION_ANCHOR_PATTERN.match(settings["viewportanchor"])
            if m:
                viewportanchor_x = float(m.group(1))
                viewportanchor_y = float(m.group(2))

        # Calculate the top-left corner (origin) of the region box.
        # Each line is ~5.33% of viewport height (LINE_HEIGHT_VH).
        height = lines * LINE_HEIGHT_VH
        # The origin is where the viewport anchor is, offset back by how far
        # the region anchor is into the box (as a fraction of box dimensions).
        origin_x = viewportanchor_x - (regionanchor_x / 100.0 * width)
        origin_y = viewportanchor_y - (regionanchor_y / 100.0 * height)

        origin = Point(
            Size(origin_x, UnitEnum.PERCENT),
            Size(origin_y, UnitEnum.PERCENT),
        )
        extent = Stretch(
            Size(width, UnitEnum.PERCENT),
            Size(height, UnitEnum.PERCENT),
        )

        return Layout(
            origin=origin,
            extent=extent
        )

    @staticmethod
    def _extract_region_id(cue_settings):
        """Extract region id from cue settings string, if present."""
        for setting in cue_settings.split():
            if setting.startswith("region:"):
                return setting[7:]
        return None


class WebVTTWriter(BaseWriter):
    HEADER = "WEBVTT\n\n"

    def write(self, caption_set, lang=None):
        """
        :type caption_set: CaptionSet
        :type lang: str
        """
        output = self.HEADER

        if caption_set.is_empty():
            return output

        caption_set = deepcopy(caption_set)

        # TODO: styles. These go into a separate CSS file, which doesn't really
        # fit the API here. Figure that out.  Though some style stuff can be
        # done in-line.  This format is a little bit crazy.

        if lang is None:
            lang = caption_set.get_languages()[0]

        self.global_layout = caption_set.get_layout_info(lang)

        captions = caption_set.get_captions(lang)

        return output + "\n".join(
            [self._convert_caption(caption_set, caption) for caption in captions]
        )

    def _timestamp(self, ts):
        td = datetime.timedelta(microseconds=ts)
        mm, ss = divmod(td.seconds, 60)
        hh, mm = divmod(mm, 60)
        s = f"{mm:02}:{ss:02}.{td.microseconds // 1000:03}"
        if hh:
            s = f"{hh:02}:{s}"
        return s

    @staticmethod
    def _convert_style_to_text_tag(style):
        if style == "italics":
            return ["<i>", "</i>"]
        elif style == "underline":
            return ["<u>", "</u>"]
        elif style == "bold":
            return ["<b>", "</b>"]
        else:
            return ["", ""]

    def _calculate_resulting_style(self, style, caption_set):
        resulting_style = {}

        style_classes = []
        if "classes" in style:
            style_classes = style["classes"]
        elif "class" in style:
            style_classes = [style["class"]]

        for style_class in style_classes:
            sub_style = caption_set.get_style(style_class).copy()
            # Recursively resolve class attributes and calculate style
            resulting_style.update(
                self._calculate_resulting_style(sub_style, caption_set)
            )

        resulting_style.update(style)

        return resulting_style

    def _convert_caption(self, caption_set, caption):
        """
        :type caption: Caption
        """
        layout_groups = self._group_cues_by_layout(caption.nodes, caption_set)

        start = self._timestamp(caption.start)
        end = self._timestamp(caption.end)
        timespan = f"{start} --> {end}"

        output = ""

        cue_style_tags = ["", ""]

        # Text styling
        style = self._calculate_resulting_style(caption.style, caption_set)
        for key, value in sorted(style.items()):
            if value:
                tags = self._convert_style_to_text_tag(key)
                cue_style_tags[0] += tags[0]
                cue_style_tags[1] = tags[1] + cue_style_tags[1]

        for cue_text, layout in layout_groups:
            if not layout:
                layout = caption.layout_info or self.global_layout
            cue_settings = self._convert_positioning(layout)
            output += timespan + cue_settings + "\n"
            output += cue_style_tags[0] + cue_text + cue_style_tags[1] + "\n"

        return output

    def _convert_positioning(self, layout):
        """
        Return WebVTT cue settings string based on layout info
        :type layout: Layout
        :rtype: str
        """
        if not layout:
            return ""

        # If it's converting from WebVTT to WebVTT, keep positioning info
        # unchanged
        if layout.webvtt_positioning:
            return f" {layout.webvtt_positioning}"

        left_offset = None
        top_offset = None
        cue_width = None

        already_relative = False
        if not self.relativize:
            if layout.is_relative():
                already_relative = True
            else:
                # There are absolute positioning values for this cue but the
                # Writer is explicitly configured not to do any relativization.
                # Ignore all positioning for this cue.
                return ""

        # Ensure that all positioning values are measured using percentage.
        # This may raise an exception if layout.is_relative() == False
        # If you want to avoid it, you have to turn off relativization by
        # initializing this Writer with relativize=False.
        if not already_relative:
            layout = layout.as_percentage_of(self.video_width, self.video_height)

        # Ensure that when there's a left offset the caption is not pushed out
        # of the screen. If the execution got this far it means origin and
        # extent are already relative by now.
        if self.fit_to_screen:
            layout = layout.fit_to_screen()

        if layout.origin:
            left_offset = layout.origin.x
            top_offset = layout.origin.y

        if layout.extent:
            cue_width = layout.extent.horizontal

        if layout.padding:
            if layout.padding.start and left_offset:
                # Since there is no padding in WebVTT, the left padding is
                # added to the total left offset (if it is defined and not
                # relative),
                left_offset += layout.padding.start
                # and removed from the total cue width
                if cue_width:
                    cue_width -= layout.padding.start
            # the right padding is cut out of the total cue width,
            if layout.padding.end and cue_width:
                cue_width -= layout.padding.end
            # the top padding is added to the top offset
            # (if it is defined and not relative)
            if layout.padding.before and top_offset:
                top_offset += layout.padding.before
            # and the bottom padding is ignored because the cue box is only as
            # long vertically as the text it contains and nothing can be cut
            # out

        if layout.alignment:
            alignment = WEBVTT_VERSION_OF.get(
                layout.alignment.horizontal, DEFAULT_ALIGN
            )
        else:
            alignment = DEFAULT_ALIGN
        cue_settings = ""

        if alignment and alignment != WEBVTT_VERSION_OF[HorizontalAlignmentEnum.CENTER]:
            # Not sure why this condition was here, maybe because center
            # alignment is applied automatically without needing to specify it
            cue_settings += f" align:{alignment}"
        if left_offset:
            cue_settings += f" position:{left_offset}"
        if top_offset:
            cue_settings += f" line:{top_offset}"
        if cue_width:
            cue_settings += f" size:{cue_width}"

        return cue_settings

    def _group_cues_by_layout(self, nodes, caption_set):
        """
        Convert a Caption's nodes to WebVTT cue or cues (depending on
        whether they have the same positioning or not).
        """
        if not nodes:
            return []

        current_layout = None

        # A list with layout groups. Since WebVTT only support positioning
        # for different cues, each layout group has to be represented in a
        # new cue with the same timing but different positioning settings.
        layout_groups = []
        # A properly encoded WebVTT string (plain unicode must be properly
        # escaped before being appended to this string)
        s = ""
        for i, node in enumerate(nodes):
            if node.type_ == CaptionNode.TEXT:
                if s and current_layout and node.layout_info != current_layout:
                    # If the positioning changes from one text node to
                    # another, a new WebVTT cue has to be created.
                    layout_groups.append((s, current_layout))
                    s = ""
                # ATTENTION: This is where the plain unicode node content is
                # finally encoded as WebVTT.
                s += self._encode_illegal_characters(node.content) or "&nbsp;"
                current_layout = node.layout_info
            elif node.type_ == CaptionNode.STYLE:
                resulting_style = self._calculate_resulting_style(
                    node.content, caption_set
                )

                styles = ["italics", "underline", "bold"]
                if not node.start:
                    styles.reverse()

                for style in styles:
                    if style in resulting_style and resulting_style[style]:
                        tags = self._convert_style_to_text_tag(style)
                        if node.start:
                            s += tags[0]
                        else:
                            s += tags[1]

                # TODO: Refactor pycaption and eliminate the concept of a
                # "Style node"
            elif node.type_ == CaptionNode.BREAK:
                if i > 0 and nodes[i - 1].type_ != CaptionNode.TEXT:
                    s += "&nbsp;"
                if i == 0:  # cue text starts with a break
                    s += "&nbsp;"
                s += "\n"

        if s:
            layout_groups.append((s, current_layout))
        return layout_groups

    def _encode_illegal_characters(self, s):
        """
        Convert cue text from plain unicode to WebVTT XML-like format
        escaping illegal characters. For a list of illegal characters see:
            - http://dev.w3.org/html5/webvtt/#dfn-webvtt-cue-text-span
        :type s: str
        """
        s = s.replace("&", "&amp;")
        s = s.replace("<", "&lt;")

        # The substring "-->" is also not allowed according to this:
        #   - http://dev.w3.org/html5/webvtt/#dfn-webvtt-cue-block
        s = s.replace("-->", "--&gt;")

        # The following characters have escaping codes for some reason, but
        # they're not illegal, so for now I'll leave this commented out so that
        # we stay as close as possible to the specification and avoid doing
        # extra stuff "just to be safe".
        # s = s.replace('>', '&gt;')
        # s = s.replace('\u200e', '&lrm;')
        # s = s.replace('\u200f', '&rlm;')
        # s = s.replace('\u00a0', '&nbsp;')
        return s



import collections

from ..base import Caption, CaptionList, CaptionNode
from ..geometry import (
    Alignment,
    HorizontalAlignmentEnum,
    Layout,
    Point,
    Size,
    UnitEnum,
    VerticalAlignmentEnum,
)
from .constants import (
    BACKGROUND_COLOR_CODES,
    COMMANDS,
    EXTENDED_CHARS,
    ITALICS_COMMANDS,
    MICROSECONDS_PER_CODEWORD,
    MID_ROW_CODES,
    PAC_BYTES_TO_POSITIONING_MAP,
    PAC_TAB_OFFSET_COMMANDS,
    PLAIN_TEXT_COMMANDS,
    STYLE_SETTING_COMMANDS,
    UNDERLINE_COMMANDS,
)

PopOnCue = collections.namedtuple("PopOnCue", "buffer, start, end")

# First two hex chars of SCC codes that produce punctuation ['.', '!', '?', ',']
_PUNCTUATION_PREFIXES = frozenset(["ae", "a1", "bf", "2c"])


class PreCaption:
    """
    The Caption class has been refactored and now its instances must be used as
    immutable objects. Some of the code in this module, however, relied on the
    fact that Caption instances were mutable. For backwards compatibility,
    therefore, this class was created to work as a mutable caption data holder
    used to eventually instantiate an actual Caption object.
    """

    def __init__(self, start=0, end=0):
        self.start = start
        self.end = end
        self.nodes = []
        self.style = {}
        self.layout_info = None

    def to_real_caption(self):
        return Caption(self.start, self.end, self.nodes, self.style, self.layout_info)


class TimingCorrectingCaptionList(list):
    """List of captions. When appending new elements, it will correct the end time
    of the last ones, so they end when the new caption gets added.

    "last ones" could mean the last caption `append`ed or all of the last
    captions with which this list was `extended`

    Also, doesn't allow Nones or empty captions
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._last_batch = ()

    def append(self, p_object):
        """When appending a new caption to the list, make sure the last one
        has an end. Also, don't add empty captions

        :type p_object: Caption | None
        """
        if p_object is None or not p_object.nodes:
            return

        self._update_last_batch(self._last_batch, p_object)

        self._last_batch = (p_object,)

        super().append(p_object)

    def extend(self, iterable):
        """Adds the elements in the iterable to the list, regarding the first
        caption's start time as the end time for the previously added
        caption(s)

        :param iterable: an iterable of Caption instances
        """
        appendable_items = [item for item in iterable if item and item.nodes]
        self._update_last_batch(self._last_batch, *appendable_items)

        self._last_batch = tuple(appendable_items)

        super().extend(appendable_items)

    @staticmethod
    def _update_last_batch(batch, *new_captions):
        """Given a batch of captions, sets their end time equal to the start
        time of the first caption in *new_captions

        The start time of the first caption in new_captions should never be 0.
        This means an invalid SCC file.

        :type batch: tuple[Caption]
        :type new_captions: tuple[Caption]
        """
        if not new_captions:
            return
        if not new_captions[0]:
            return
        if not new_captions[0].nodes:
            return

        new_caption = new_captions[0]

        if batch and (
            batch[-1].end == 0
            or new_caption.start - batch[-1].end < 5 * MICROSECONDS_PER_CODEWORD + 1
        ):
            for caption in batch:
                caption.end = new_caption.start


class NotifyingDict(dict):
    """Dictionary-like object, that treats one key as 'active',
    and notifies observers if the active key changed
    """

    # Need an unhashable object as initial value for the active key.
    # That way we're sure this was never a key in the dict.
    _guard = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.active_key = self._guard
        self.observers = []

    def set_active(self, key):
        """Sets the active key

        :param key: any hashable object
        """
        if key not in self:
            raise ValueError("No such key present")

        # Notify observers of the change
        if key != self.active_key:
            for observer in self.observers:
                observer(self.active_key, key)

        self.active_key = key

    def get_active(self):
        """Returns the value corresponding to the active key"""
        if self.active_key is self._guard:
            raise KeyError("No active key set")

        return self[self.active_key]

    def add_change_observer(self, observer):
        """Receives a callable function, which it will call if the active
        element changes.

        The observer will receive 2 positional arguments: the old and new key

        :param observer: any callable that can be called with 2 positional
            arguments
        """
        if not callable(observer):
            raise TypeError("The observer should be callable")

        self.observers.append(observer)


class CaptionCreator:
    """Creates and maintains a collection of Captions"""

    def __init__(self):
        self._collection = TimingCorrectingCaptionList()

        # subset of self._collection;
        # captions here will be susceptible to time corrections
        self._still_editing = []

    def correct_last_timing(self, end_time, force=False):
        """Called to set the time on the last Caption(s) stored with no end
        time

        :type force: bool
        :param force: Set the end time even if there's already an end time

        :type end_time: float
        :param end_time: microseconds; the end of the caption;
        """
        if not self._still_editing:
            return

        if force:
            # Select all last captions
            captions_to_correct = self._still_editing
        elif self._still_editing[-1].end == 0:
            # Only select the last captions if they haven't gotten their
            # end time set yet
            captions_to_correct = self._still_editing
        else:
            return

        for caption in captions_to_correct:
            caption.end = end_time

    def create_and_store(self, node_buffer, start, end=0):
        """Interpreter method, will convert the buffer into one or more Caption
        objects, storing them internally.

        This method relies on the InstructionNodeCreator's ability to generate
        InstructionNodes properly, so at this point we can convert
        _InstructionNodes nodes almost 1:1 to CaptionNodes

        :type node_buffer: InstructionNodeCreator

        :type start: float
        :param start: the start time in microseconds
        :type end: float
        :param end: the end time in microseconds
        """
        if node_buffer.is_empty():
            return

        caption = PreCaption()
        caption.start = start
        caption.end = end
        self._still_editing = [caption]

        for instruction in node_buffer:
            # skip empty elements
            if instruction.is_empty():
                continue

            elif instruction.requires_repositioning():
                caption = PreCaption()
                caption.start = start
                caption.end = end
                self._still_editing.append(caption)

            # handle line breaks
            elif instruction.is_explicit_break():
                caption.nodes.append(
                    CaptionNode.create_break(
                        layout_info=_get_layout_from_tuple(instruction.position)
                    )
                )

            # handle open italics
            elif instruction.sets_italics_on():
                caption.nodes.append(
                    CaptionNode.create_style(
                        True,
                        {"italics": True},
                        layout_info=_get_layout_from_tuple(instruction.position),
                    )
                )

            # handle clone italics
            elif instruction.sets_italics_off():
                caption.nodes.append(
                    CaptionNode.create_style(
                        False,
                        {"italics": True},
                        layout_info=_get_layout_from_tuple(instruction.position),
                    )
                )

            # handle text
            elif instruction.is_text_node():
                layout_info = _get_layout_from_tuple(instruction.position)
                caption.nodes.append(
                    CaptionNode.create_text(
                        text=instruction.text,
                        layout_info=layout_info,
                        position=instruction.position,
                    )
                )
                caption.layout_info = layout_info

        self._collection.extend(self._still_editing)

    def get_all(self):
        """Returns the Caption collection as a CaptionList

        :rtype: CaptionList
        """
        caption_list = CaptionList()
        for precap in self._collection:
            caption_list.append(precap.to_real_caption())
        return caption_list


class InstructionNodeCreator:
    """Creates _InstructionNode instances from characters and commands, storing
    them internally
    """

    def __init__(self, collection=None, position_tracker=None):
        """
        :param collection: an optional collection of nodes

        :param position_tracker:
        :return:
        """
        if not collection:
            self._collection = []
        else:
            self._collection = collection

        self.last_style = (
            None  # can be italic on or italic off as we only support italics
        )
        self._position_tracer = position_tracker

    def is_empty(self):
        """Whether any text was added to the buffer"""
        return not any(element.text for element in self._collection)

    def add_chars(self, *chars):
        """Adds characters to a text node (last text node, or a new one)

        :param chars: tuple containing text (Unicode string)
        """
        if not chars:
            return

        current_position = self._position_tracer.get_current_position()

        # get or create a usable node
        if (
            self._collection
            and self._collection[-1].is_text_node()
            and not self._position_tracer.is_repositioning_required()
        ):
            node = self._collection[-1]
        else:
            # create first node
            node = _InstructionNode(position=current_position)
            self._collection.append(node)

        # handle line break(s) - may be multiple for multi-row jumps
        if self._position_tracer.is_linebreak_required():
            for _ in range(self._position_tracer._breaks_required):
                self._collection.append(
                    _InstructionNode.create_break(position=current_position)
                )
            self._position_tracer.acknowledge_linebreak_consumed()
            node = _InstructionNode.create_text(current_position)
            self._collection.append(node)
            if self._position_tracer.is_repositioning_required():
                # it means we have a reposition command which was not followed by
                # any text, so we just ignore it and break
                self._position_tracer.acknowledge_position_changed()

        # handle completely new positioning
        elif self._position_tracer.is_repositioning_required():
            self._collection.append(
                _InstructionNode.create_repositioning_command(current_position)
            )
            node = _InstructionNode.create_text(current_position)
            self._collection.append(node)
            self._position_tracer.acknowledge_position_changed()

        node.add_chars(*chars)

    @staticmethod
    def get_style_for_command(command):
        if command in ITALICS_COMMANDS:
            return "italic"
        elif command in UNDERLINE_COMMANDS:
            return "underline"
        else:
            # as we only check STYLE_SETTING_COMMANDS,
            # only remaining possibility is plain text
            return "plaintext"

    def interpret_command(self, command, next_command=None):
        """Given a command determines whether to turn italics on or off,
        or to set the positioning

        This is mostly used to convert from the legacy-style commands

        :type command: str
        or a PAC_TAB_OFFSET_COMMANDS
        :type next_command: the command that follows next
        """
        self._update_positioning(command)

        if command == "94a1":
            self.handle_backspace("94a1")

        if command in BACKGROUND_COLOR_CODES:
            # Since these codes are optional, they must be preceded
            # with the space character (20h),
            # which will be deleted when the code is applied.
            # ex: 2080 97ad 94a1
            if (
                len(self._collection) > 0
                and self._collection[-1].is_text_node()
                and self._collection[-1].text[-1].isspace()
            ):
                self._collection[-1].text = self._collection[-1].text[:-1]

        if command in STYLE_SETTING_COMMANDS:
            current_position = self._position_tracer.get_current_position()
            # which style is command setting
            command_style = self.get_style_for_command(command)
            if command_style == "italic":
                if self.last_style is None or self.last_style == "italics off":
                    #  if we don't have any style yet, or we have a closed italics tag
                    #  it should open italic tag
                    #  if break is required, break then add style tag
                    if self._position_tracer.is_linebreak_required():
                        for _ in range(self._position_tracer._breaks_required):
                            self._collection.append(
                                _InstructionNode.create_break(position=current_position)
                            )
                        self._position_tracer.acknowledge_linebreak_consumed()
                    self._collection.append(
                        _InstructionNode.create_italics_style(current_position)
                    )
                    self.last_style = "italics on"
            else:
                # command sets a different style (underline, plain)
                # so we need to close italics if we have an open italics tag
                # otherwise we ignore it
                # if break is required, add style tag then break
                if self.last_style == "italics on":
                    self._collection.append(
                        _InstructionNode.create_italics_style(
                            self._position_tracer.get_current_position(), turn_on=False
                        )
                    )
                    self.last_style = "italics off"
                    if self._position_tracer.is_linebreak_required():
                        for _ in range(self._position_tracer._breaks_required):
                            self._collection.append(
                                _InstructionNode.create_break(position=current_position)
                            )
                        self._position_tracer.acknowledge_linebreak_consumed()

        #  handle mid-row codes that follows a text node
        #  don't add space if the next command adds one of
        #  ['.', '!', '?', ',']
        next_is_punctuation = next_command and next_command[:2] in _PUNCTUATION_PREFIXES
        prev_text_node = self.get_previous_text_node()
        prev_node_is_break = prev_text_node is not None and any(
            x.is_explicit_break()
            for x in self._collection[self._collection.index(prev_text_node) :]
        )
        if (
            command in MID_ROW_CODES
            and prev_text_node
            and not prev_node_is_break
            and not prev_text_node.text[-1].isspace()
            and command not in PAC_TAB_OFFSET_COMMANDS
            and not next_is_punctuation
        ):
            if self.last_style == "italics off":
                # need to open italics tag, add a space
                # to the beginning of the next text node
                self.add_chars(" ")
            else:
                # italics on
                # need to close italics tag, add a space
                # to the end of the previous text node
                prev_text_node.text = prev_text_node.text + " "

    def _update_positioning(self, command):
        """Sets the positioning information to use for the next nodes

        :type command: str
        """
        is_offset = False
        if command in PAC_TAB_OFFSET_COMMANDS:
            prev_positioning = self._position_tracer.default
            tab_offset = PAC_TAB_OFFSET_COMMANDS[command]
            positioning = (prev_positioning[0], prev_positioning[1] + tab_offset)
            is_offset = True
        else:
            first, second = command[:2], command[2:]
            try:
                # is PAC
                positioning = PAC_BYTES_TO_POSITIONING_MAP[first][second]
            except KeyError:
                # if not PAC or OFFSET we're not changing position
                return
        offset_after_break = is_offset and self.has_break_before(self._collection)
        if not offset_after_break:
            # Tab offsets after line breaks will be ignored to avoid repositioning
            self._position_tracer.update_positioning(positioning)

    def __iter__(self):
        return iter(_format_italics(self._collection))

    @staticmethod
    def has_break_before(collection):
        if len(collection) == 0:
            return False
        for element in collection[::-1]:
            if element._type == _InstructionNode.TEXT:
                return False
            elif element._type == _InstructionNode.BREAK:
                return True
        return False

    @classmethod
    def from_list(cls, stash_list, position_tracker):
        """Having received a list of instances of this class, creates a new
        instance that contains all the nodes of the previous instances
        (basically concatenates the many stashes into one)

        :type stash_list: list[InstructionNodeCreator]
        :param stash_list: a list of instances of this class

        :type position_tracker: .state_machines.DefaultProvidingPositionTracker
        :param position_tracker: state machine to be interrogated about the
            positioning when creating a node

        :rtype: InstructionNodeCreator
        """
        instance = cls(position_tracker=position_tracker)
        new_collection = instance._collection

        for idx, stash in enumerate(stash_list):
            new_collection.extend(stash._collection)

            # use space to separate the stashes, but don't add final space
            if idx < len(stash_list) - 1:
                try:
                    instance._collection[-1].add_chars(" ")
                except AttributeError:
                    pass

        return instance

    def handle_backspace(self, word):
        """
        Move cursor back one position and delete that character
        """
        node = self.get_previous_text_node()
        # in case of no previous text nodes or
        # if the backspace is required while no character
        # do nothing
        if node is None:
            return
        last_char = node.text[-1]
        delete_previous_condition = (
            word in EXTENDED_CHARS and last_char not in EXTENDED_CHARS.values()
        ) or word == "94a1"
        # in case extended char, perform backspace
        # only if the previous character in not also extended
        if delete_previous_condition:
            node.text = node.text[:-1]

    def get_previous_text_node(self):
        for node in self._collection[::-1]:
            if node.is_text_node() and node.text:
                return node
        return None


def _get_layout_from_tuple(position_tuple):
    """Create a Layout object from the positioning information given

    The row can have a value from 1 to 15 inclusive. (vertical positioning)
    The column can have a value from 0 to 31 inclusive. (horizontal)

    :param position_tuple: a tuple of ints (row, col)
    :type position_tuple: tuple
    :rtype: Layout
    """
    if not position_tuple:
        return None

    row, column = position_tuple

    # Horizontal safe area between 10% and 90%
    horizontal = Size(80 * column / 32.0 + 10, UnitEnum.PERCENT)
    # Vertical safe area between 5% and 95%
    vertical = Size(90 * (row - 1) / 15.0 + 5, UnitEnum.PERCENT)
    return Layout(
        origin=Point(horizontal, vertical),
        alignment=Alignment(HorizontalAlignmentEnum.LEFT, VerticalAlignmentEnum.TOP),
    )


class _InstructionNode:
    """Value object, that can contain text information, or interpretable
    commands (such as explicit line breaks or turning italics on/off).

    These nodes will be aggregated into a RepresentableNode, which will then
    be easily converted to a CaptionNode.
    """

    TEXT = 0
    BREAK = 1
    ITALICS_ON = 2
    ITALICS_OFF = 3
    CHANGE_POSITION = 4

    def __init__(self, text=None, position=None, type_=0):
        """
        :type text: str
        :param position: a tuple of ints (row, column)
        :param type_: self.TEXT | self.BREAK | self.ITALICS
        :type type_: int
        """
        self.text = text
        self.position = position
        self._type = type_

    def add_chars(self, *args):
        """This being a text node, add characters to it.
        :param args:
        :type args: tuple[str]
        :return:
        """
        if self.text is None:
            self.text = ""

        self.text += "".join(args)

    def is_text_node(self):
        """
        :rtype: bool
        """
        return self._type == self.TEXT

    def is_empty(self):
        """
        :rtype: bool
        """
        if self._type == self.TEXT:
            return not self.text

        return False

    def is_explicit_break(self):
        """
        :rtype: bool
        """
        return self._type == self.BREAK

    def sets_italics_on(self):
        """
        :rtype: bool
        """
        return self._type == self.ITALICS_ON

    def sets_italics_off(self):
        """
        :rtype: bool
        """
        return self._type == self.ITALICS_OFF

    def is_italics_node(self):
        """
        :rtype: bool
        """
        return self._type in (self.ITALICS_OFF, self.ITALICS_ON)

    def requires_repositioning(self):
        """Whether the node must be interpreted as a change in positioning

        :rtype: bool
        """
        return self._type == self.CHANGE_POSITION

    def get_text(self):
        """A little legacy code."""
        return " ".join(self.text.split())

    @classmethod
    def create_break(cls, position):
        """Create a node, interpretable as an explicit line break

        :type position: tuple[int]
        :param position: a tuple (row, col) containing the positioning info

        :rtype: _InstructionNode
        """
        return cls(type_=cls.BREAK, position=position)

    @classmethod
    def create_text(cls, position, *chars):
        """Create a node interpretable as text

        :type position: tuple[int]
        :param position: a tuple (row, col) to mark the positioning

        :type chars: tuple[str]
        :param chars: characters to add to the text

        :rtype: _InstructionNode
        """
        return cls("".join(chars), position=position)

    @classmethod
    def create_italics_style(cls, position, turn_on=True):
        """Create a node, interpretable as a command to switch italics on/off

        :type position: tuple[int]
        :param position: a tuple (row, col) to mark the positioning

        :type turn_on: bool
        :param turn_on: whether to turn the italics on or off

        :rtype: _InstructionNode
        """
        return cls(
            position=position, type_=cls.ITALICS_ON if turn_on else cls.ITALICS_OFF
        )

    @classmethod
    def create_repositioning_command(cls, position=None):
        """Create node interpretable as a command to change the current
        position

        :type position:
        """
        return cls(type_=cls.CHANGE_POSITION, position=position)

    def __repr__(self):  # pragma: no cover
        if self._type == self.BREAK:
            extra = "BR"
        elif self._type == self.TEXT:
            extra = f'"{self.text}"'
        elif self._type in (self.ITALICS_ON, self.ITALICS_OFF):
            extra = "italics {}".format(
                "on" if self._type == self.ITALICS_ON else "off"
            )
        else:
            extra = "change position"

        return f"<INode: {extra} >"


def _format_italics(collection):
    """Given a raw list of _InstructionNodes, returns a new equivalent list
    where all the italics nodes properly close and open.

    The list is equivalent in the sense that the SCC commands that would have
    generated the output list, would have had the exact same visual effect
    as the ones that generated the output, as far as italics are concerned.

    This is useful because the raw commands read from the SCC can't be used
    the way they are by the writers for the other formats. Those other writers
    require the list of CaptionNodes to be formatted in a certain way.

    Note: Using state machines to manage the italics didn't work well because
    we're using state machines already to track the position, and their
    interactions got crazy.

    :type collection: list[_InstructionNode]
    :rtype: list[_InstructionNode]
    """
    new_collection = _skip_initial_italics_off_nodes(collection)

    new_collection = _skip_empty_text_nodes(new_collection)

    # after this step we're guaranteed a proper ordering of the nodes
    new_collection = _skip_redundant_italics_nodes(new_collection)

    # after this, we're guaranteed that the italics are properly contained
    # within their context
    new_collection = _close_italics_before_repositioning(new_collection)

    # all nodes will be closed after this step
    new_collection = _ensure_final_italics_node_closes(new_collection)

    # removes pairs of italics nodes that don't do anything noticeable
    new_collection = _remove_noop_italics(new_collection)

    # remove spaces to the end of the lines
    new_collection = _remove_spaces_at_end_of_the_line(new_collection)

    return new_collection


def _remove_spaces_at_end_of_the_line(collection):
    if not collection:
        return collection
    for idx, node in enumerate(collection):
        if (
            idx > 0
            and node._type == _InstructionNode.BREAK
            and collection[idx - 1].is_text_node()
            and collection[idx - 1].text
        ):
            collection[idx - 1].text = collection[idx - 1].text.rstrip()
    # handle last node
    if collection[-1].is_text_node():
        collection[-1].text = collection[-1].text.rstrip()
    return collection


def _remove_noop_on_off_italics(collection):
    """Return an equivalent list to `collection`. It removes the italics node
     pairs that don't surround text nodes, if those nodes are in the order:
     on, off

    :type collection: list[_InstructionNode]
    :rtype: list[_InstructionNode]
    """
    new_collection = []
    to_commit = None

    for node in collection:
        if node.is_italics_node() and node.sets_italics_on():
            to_commit = node
            continue

        elif node.is_italics_node() and node.sets_italics_off():
            if to_commit:
                to_commit = None
                continue
        else:
            if to_commit:
                new_collection.append(to_commit)
                to_commit = None

        new_collection.append(node)

    return new_collection


def _remove_noop_off_on_italics(collection):
    """Removes pairs of off-on italics nodes, that don't surround any other
    node

    :type collection: list[_InstructionNode]
    :return: list[_InstructionNode]
    """
    new_collection = []
    to_commit = None

    for node in collection:
        if node.is_italics_node() and node.sets_italics_off():
            to_commit = node
            continue

        elif node.is_italics_node() and node.sets_italics_on():
            if to_commit:
                to_commit = None
                continue
        else:
            if to_commit:
                new_collection.append(to_commit)
                to_commit = None

        new_collection.append(node)

    if to_commit:
        new_collection.append(to_commit)

    return new_collection


def _remove_noop_italics(collection):
    """Return an equivalent list to `collection`. It removes the italics node
     pairs that don't surround text nodes

    :type collection: list[_InstructionNode]
    :rtype: list[_InstructionNode]
    """
    new_collection = _remove_noop_on_off_italics(collection)

    new_collection = _remove_noop_off_on_italics(new_collection)

    return new_collection


def _skip_initial_italics_off_nodes(collection):
    """Return a collection like the one given, but without the
    initial <Italics OFF> nodes

    :type collection: list[_InstructionNode]
    :rtype: list[_InstructionNode]
    """
    new_collection = []
    can_add_italics_off_nodes = False

    for node in collection:
        if node.is_italics_node():
            if node.sets_italics_on():
                can_add_italics_off_nodes = True
                new_collection.append(node)
            elif can_add_italics_off_nodes:
                new_collection.append(node)
        else:
            new_collection.append(node)

    return new_collection


def _skip_empty_text_nodes(collection):
    """Return an iterable containing all the nodes in the previous
    collection except for the empty text nodes

    :type collection: list[_InstructionNode]
    :rtype: list[_InstructionNode]
    """
    return [
        node for node in collection if not (node.is_text_node() and node.is_empty())
    ]


def _skip_redundant_italics_nodes(collection):
    """Return a list where the <Italics On> nodes only appear after
    <Italics OFF>, and vice versa. This ignores the other node types, and
    only removes redundant italic nodes

    :type collection: list[_InstructionNode]
    :rtype: list[_InstructionNode]
    """
    new_collection = []
    state = None

    for node in collection:
        if node.is_italics_node():
            if state is None:
                state = node.sets_italics_on()
                if node.sets_italics_on():
                    new_collection.append(node)
                continue
            # skip the nodes that are like the previous
            if node.sets_italics_on() is state:
                continue
            else:
                state = node.sets_italics_on()
        new_collection.append(node)

    return new_collection


def _close_italics_before_repositioning(collection):
    """Make sure that for every opened italic node, there's a corresponding
    closing node.

    Will insert a closing italic node, before each repositioning node

    :type collection: list[_InstructionNode]
    :rtype: list[_InstructionNode]
    """
    new_collection = []

    italics_on = False
    last_italics_on_node = None

    for node in collection:
        if node.is_italics_node() and node.sets_italics_on():
            italics_on = True
            last_italics_on_node = node
        if node.is_italics_node() and node.sets_italics_off():
            italics_on = False
        if node.requires_repositioning() and italics_on:
            # Append an italics closing node before the position change
            new_collection.append(
                _InstructionNode.create_italics_style(
                    # The position info of this new node should be the same
                    position=last_italics_on_node.position,
                    turn_on=False,
                )
            )
            new_collection.append(node)
            # Append an italics opening node after the positioning change
            new_collection.append(
                _InstructionNode.create_italics_style(position=node.position)
            )
            continue

        new_collection.append(node)

    return new_collection


def _ensure_final_italics_node_closes(collection):
    """The final italics command needs to be closed

    :type collection: list[_InstructionNode]
    :rtype: list[_InstructionNode]
    """
    new_collection = list(collection)

    italics_on = False
    last_italics_on_node = None

    for node in collection:
        if node.is_italics_node() and node.sets_italics_on():
            italics_on = True
            last_italics_on_node = node
        if node.is_italics_node() and node.sets_italics_off():
            italics_on = False

    if italics_on:
        new_collection.append(
            _InstructionNode.create_italics_style(
                position=last_italics_on_node.position, turn_on=False
            )
        )
    return new_collection
